@@ -9,45 +9,135 @@ use crate::commands::CreateExe;
 #[cfg(feature = "wast")]
 use crate::commands::Wast;
 use crate::commands::{
-    Add, Cache, Config, Init, Inspect, List, Login, Publish, Run, RunUnstable, SelfUpdate,
-    Validate, Whoami,
+    Add, Cache, Config, Init, Inspect, Login, Publish, Run, SelfUpdate, Validate, Whoami,
 };
 #[cfg(feature = "static-artifact-create")]
 use crate::commands::{CreateObj, GenCHeader};
 use crate::error::PrettyError;
-use clap::{CommandFactory, ErrorKind, Parser};
+use clap::{CommandFactory, Parser};
+use wasmer_deploy_cli::cmd::CliCommand;
+
+/// The main function for the Wasmer CLI tool.
+pub fn wasmer_main() {
+    // We allow windows to print properly colors
+    #[cfg(windows)]
+    colored::control::set_virtual_terminal(true).unwrap();
+
+    PrettyError::report(wasmer_main_inner())
+}
+
+fn wasmer_main_inner() -> Result<(), anyhow::Error> {
+    if is_binfmt_interpreter() {
+        Run::from_binfmt_args().execute(crate::logging::Output::default());
+    }
+
+    match Args::try_parse() {
+        Ok(args) => args.execute(),
+        Err(e) => {
+            let might_be_wasmer_run = matches!(
+                e.kind(),
+                clap::error::ErrorKind::InvalidSubcommand | clap::error::ErrorKind::UnknownArgument
+            );
+
+            if might_be_wasmer_run {
+                if let Ok(run) = Run::try_parse() {
+                    // Try to parse the command using the `wasmer some/package`
+                    // shorthand. Note that this has discoverability issues
+                    // because it's not shown as part of the main argument
+                    // parser's help, but that's fine.
+                    let output = crate::logging::Output::default();
+                    output.initialize_logging();
+                    run.execute(output);
+                }
+            }
+
+            e.exit();
+        }
+    }
+}
+
+/// Command-line arguments for the Wasmer CLI.
+#[derive(Parser, Debug)]
+#[clap(author, version)]
+#[clap(disable_version_flag = true)] // handled manually
+#[cfg_attr(feature = "headless", clap(
+    name = "wasmer-headless",
+    about = concat!("wasmer-headless ", env!("CARGO_PKG_VERSION")),
+))]
+#[cfg_attr(not(feature = "headless"), clap(
+    name = "wasmer",
+    about = concat!("wasmer ", env!("CARGO_PKG_VERSION")),
+))]
+pub struct Args {
+    /// Print version info and exit.
+    #[clap(short = 'V', long)]
+    version: bool,
+    #[clap(flatten)]
+    output: crate::logging::Output,
+    #[clap(subcommand)]
+    cmd: Option<Cmd>,
+}
+
+impl Args {
+    fn execute(self) -> Result<(), anyhow::Error> {
+        let Args {
+            cmd,
+            version,
+            output,
+        } = self;
+
+        output.initialize_logging();
+
+        if version {
+            return print_version(output.is_verbose());
+        }
+
+        match cmd {
+            Some(Cmd::Run(options)) => options.execute(output),
+            Some(Cmd::SelfUpdate(options)) => options.execute(),
+            Some(Cmd::Cache(cache)) => cache.execute(),
+            Some(Cmd::Validate(validate)) => validate.execute(),
+            #[cfg(feature = "compiler")]
+            Some(Cmd::Compile(compile)) => compile.execute(),
+            #[cfg(any(feature = "static-artifact-create", feature = "wasmer-artifact-create"))]
+            Some(Cmd::CreateExe(create_exe)) => create_exe.execute(),
+            #[cfg(feature = "static-artifact-create")]
+            Some(Cmd::CreateObj(create_obj)) => create_obj.execute(),
+            Some(Cmd::Config(config)) => config.execute(),
+            Some(Cmd::Inspect(inspect)) => inspect.execute(),
+            Some(Cmd::Init(init)) => init.execute(),
+            Some(Cmd::Login(login)) => login.execute(),
+            Some(Cmd::Publish(publish)) => publish.execute(),
+            #[cfg(feature = "static-artifact-create")]
+            Some(Cmd::GenCHeader(gen_heder)) => gen_heder.execute(),
+            #[cfg(feature = "wast")]
+            Some(Cmd::Wast(wast)) => wast.execute(),
+            #[cfg(target_os = "linux")]
+            Some(Cmd::Binfmt(binfmt)) => binfmt.execute(),
+            Some(Cmd::Whoami(whoami)) => whoami.execute(),
+            Some(Cmd::Add(install)) => install.execute(),
+
+            // Deploy commands.
+            Some(Cmd::Deploy(c)) => c.run(),
+            Some(Cmd::App(apps)) => apps.run(),
+            Some(Cmd::Ssh(ssh)) => ssh.run(),
+            Some(Cmd::Namespace(namespace)) => namespace.run(),
+            None => {
+                Args::command().print_long_help()?;
+                // Note: clap uses an exit code of 2 when CLI parsing fails
+                std::process::exit(2);
+            }
+        }
+    }
+}
 
 #[derive(Parser, Debug)]
-#[cfg_attr(
-    not(feature = "headless"),
-    clap(
-        name = "wasmer",
-        about = "WebAssembly standalone runtime.",
-        version,
-        author
-    )
-)]
-#[cfg_attr(
-    feature = "headless",
-    clap(
-        name = "wasmer-headless",
-        about = "WebAssembly standalone runtime (headless).",
-        version,
-        author
-    )
-)]
 /// The options for the wasmer Command Line Interface
-enum WasmerCLIOptions {
-    /// List all locally installed packages
-    List(List),
-
-    /// Run a WebAssembly file. Formats accepted: wasm, wat
-    Run(Run),
-
-    /// Login into a wapm.io-like registry
+enum Cmd {
+    /// Login into a wasmer.io-like registry
     Login(Login),
 
-    /// Login into a wapm.io-like registry
+    /// Login into a wasmer.io-like registry
     #[clap(name = "publish")]
     Publish(Publish),
 
@@ -159,151 +249,72 @@ enum WasmerCLIOptions {
     /// Shows the current logged in user for the current active registry
     Whoami(Whoami),
 
-    /// Add a WAPM package's bindings to your application.
+    /// Add a Wasmer package's bindings to your application.
     Add(Add),
 
-    /// (unstable) Run a WebAssembly file or WEBC container.
-    RunUnstable(RunUnstable),
+    /// Run a WebAssembly file or Wasmer container.
+    #[clap(alias = "run-unstable")]
+    Run(Run),
+
+    // DEPLOY commands
+    /// Deploy apps to the Wasmer Edge.
+    Deploy(wasmer_deploy_cli::cmd::deploy::CmdDeploy),
+
+    /// Manage deployed apps.
+    #[clap(subcommand, alias = "apps")]
+    App(wasmer_deploy_cli::cmd::app::CmdApp),
+
+    /// Create a dynamic on the Deploy Edge, and connect to it through SSH.
+    Ssh(wasmer_deploy_cli::cmd::ssh::CmdSsh),
+
+    /// Manage Wasmer namespaces.
+    #[clap(subcommand, alias = "namespaces")]
+    Namespace(wasmer_deploy_cli::cmd::namespace::CmdNamespace),
 }
 
-impl WasmerCLIOptions {
-    fn execute(&self) -> Result<(), anyhow::Error> {
-        match self {
-            Self::Run(options) => options.execute(),
-            Self::SelfUpdate(options) => options.execute(),
-            Self::Cache(cache) => cache.execute(),
-            Self::Validate(validate) => validate.execute(),
-            #[cfg(feature = "compiler")]
-            Self::Compile(compile) => compile.execute(),
-            #[cfg(any(feature = "static-artifact-create", feature = "wasmer-artifact-create"))]
-            Self::CreateExe(create_exe) => create_exe.execute(),
-            #[cfg(feature = "static-artifact-create")]
-            Self::CreateObj(create_obj) => create_obj.execute(),
-            Self::Config(config) => config.execute(),
-            Self::Inspect(inspect) => inspect.execute(),
-            Self::Init(init) => init.execute(),
-            Self::List(list) => list.execute(),
-            Self::Login(login) => login.execute(),
-            Self::Publish(publish) => publish.execute(),
-            #[cfg(feature = "static-artifact-create")]
-            Self::GenCHeader(gen_heder) => gen_heder.execute(),
-            #[cfg(feature = "wast")]
-            Self::Wast(wast) => wast.execute(),
-            #[cfg(target_os = "linux")]
-            Self::Binfmt(binfmt) => binfmt.execute(),
-            Self::Whoami(whoami) => whoami.execute(),
-            Self::Add(install) => install.execute(),
-            Self::RunUnstable(run2) => run2.execute(),
+fn is_binfmt_interpreter() -> bool {
+    cfg_if::cfg_if! {
+        if #[cfg(target_os = "linux")] {
+            // Note: we'll be invoked by the kernel as Binfmt::FILENAME
+            let binary_path = match std::env::args_os().next() {
+                Some(path) => std::path::PathBuf::from(path),
+                None => return false,
+            };
+            binary_path.file_name().and_then(|f| f.to_str()) == Some(Binfmt::FILENAME)
+        } else {
+            false
         }
     }
 }
 
-/// The main function for the Wasmer CLI tool.
-pub fn wasmer_main() {
-    // We allow windows to print properly colors
-    #[cfg(windows)]
-    colored::control::set_virtual_terminal(true).unwrap();
-
-    PrettyError::report(wasmer_main_inner())
-}
-
-fn wasmer_main_inner() -> Result<(), anyhow::Error> {
-    // We try to run wasmer with the normal arguments.
-    // Eg. `wasmer <SUBCOMMAND>`
-    // In case that fails, we fallback trying the Run subcommand directly.
-    // Eg. `wasmer myfile.wasm --dir=.`
-    //
-    // In case we've been run as wasmer-binfmt-interpreter myfile.wasm args,
-    // we assume that we're registered via binfmt_misc
-    let args = std::env::args().collect::<Vec<_>>();
-    let binpath = args.get(0).map(|s| s.as_ref()).unwrap_or("");
-
-    let firstarg = args.get(1).map(|s| s.as_str());
-    let secondarg = args.get(2).map(|s| s.as_str());
-
-    match (firstarg, secondarg) {
-        (None, _) | (Some("help"), _) | (Some("--help"), _) => {
-            return print_help(true);
-        }
-        (Some("-h"), _) => {
-            return print_help(false);
-        }
-        (Some("-vV"), _)
-        | (Some("version"), Some("--verbose"))
-        | (Some("--version"), Some("--verbose")) => {
-            return print_version(true);
-        }
-
-        (Some("-v"), _) | (Some("-V"), _) | (Some("version"), _) | (Some("--version"), _) => {
-            return print_version(false);
-        }
-        _ => {}
-    }
-
-    let command = args.get(1);
-    let options = if cfg!(target_os = "linux") && binpath.ends_with("wasmer-binfmt-interpreter") {
-        WasmerCLIOptions::Run(Run::from_binfmt_args())
-    } else {
-        match command.unwrap_or(&"".to_string()).as_ref() {
-            "add" | "cache" | "compile" | "config" | "create-obj" | "create-exe" | "help"
-            | "gen-c-header" | "inspect" | "init" | "run" | "run-unstable" | "self-update"
-            | "validate" | "wast" | "binfmt" | "list" | "login" | "publish" => {
-                WasmerCLIOptions::parse()
-            }
-            _ => {
-                WasmerCLIOptions::try_parse_from(args.iter()).unwrap_or_else(|e| {
-                    match e.kind() {
-                        // This fixes a issue that:
-                        // 1. Shows the version twice when doing `wasmer -V`
-                        // 2. Shows the run help (instead of normal help) when doing `wasmer --help`
-                        ErrorKind::DisplayVersion | ErrorKind::DisplayHelp => e.exit(),
-                        _ => WasmerCLIOptions::Run(Run::parse()),
-                    }
-                })
-            }
-        }
-    };
-
-    options.execute()
-}
-
-fn print_help(verbose: bool) -> Result<(), anyhow::Error> {
-    let mut cmd = WasmerCLIOptions::command();
-    if verbose {
-        let _ = cmd.print_long_help();
-    } else {
-        let _ = cmd.print_help();
-    }
-    Ok(())
-}
-
-#[allow(unused_mut, clippy::vec_init_then_push)]
 fn print_version(verbose: bool) -> Result<(), anyhow::Error> {
     if !verbose {
         println!("wasmer {}", env!("CARGO_PKG_VERSION"));
-    } else {
-        println!(
-            "wasmer {} ({} {})",
-            env!("CARGO_PKG_VERSION"),
-            env!("WASMER_BUILD_GIT_HASH_SHORT"),
-            env!("WASMER_BUILD_DATE")
-        );
-        println!("binary: {}", env!("CARGO_PKG_NAME"));
-        println!("commit-hash: {}", env!("WASMER_BUILD_GIT_HASH"));
-        println!("commit-date: {}", env!("WASMER_BUILD_DATE"));
-        println!("host: {}", target_lexicon::HOST);
-        println!("compiler: {}", {
-            let mut s = Vec::<&'static str>::new();
-
-            #[cfg(feature = "singlepass")]
-            s.push("singlepass");
-            #[cfg(feature = "cranelift")]
-            s.push("cranelift");
-            #[cfg(feature = "llvm")]
-            s.push("llvm");
-
-            s.join(",")
-        });
+        return Ok(());
     }
+
+    println!(
+        "wasmer {} ({} {})",
+        env!("CARGO_PKG_VERSION"),
+        env!("WASMER_BUILD_GIT_HASH_SHORT"),
+        env!("WASMER_BUILD_DATE")
+    );
+    println!("binary: {}", env!("CARGO_PKG_NAME"));
+    println!("commit-hash: {}", env!("WASMER_BUILD_GIT_HASH"));
+    println!("commit-date: {}", env!("WASMER_BUILD_DATE"));
+    println!("host: {}", target_lexicon::HOST);
+
+    let mut compilers = Vec::<&'static str>::new();
+    if cfg!(feature = "singlepass") {
+        compilers.push("singlepass");
+    }
+    if cfg!(feature = "cranelift") {
+        compilers.push("cranelift");
+    }
+    if cfg!(feature = "llvm") {
+        compilers.push("llvm");
+    }
+    println!("compiler: {}", compilers.join(","));
+
     Ok(())
 }
