@@ -20,9 +20,7 @@ use indicatif::{MultiProgress, ProgressBar};
 use once_cell::sync::Lazy;
 use sha2::{Digest, Sha256};
 use tempfile::NamedTempFile;
-use tokio::runtime::Handle;
 use url::Url;
-use wapm_targz_to_pirita::{webc::v1::DirOrFile, FileMap, TransformManifestFunctions};
 use wasmer::{
     DeserializeError, Engine, Function, Imports, Instance, Module, Store, Type, TypedFunction,
     Value,
@@ -37,6 +35,7 @@ use wasmer_wasix::{
         module_cache::{CacheError, ModuleHash},
         package_loader::PackageLoader,
         resolver::{PackageSpecifier, QueryError},
+        task_manager::VirtualTaskManagerExt,
     },
     WasiError,
 };
@@ -104,21 +103,22 @@ impl Run {
             wasmer_vm::set_stack_size(self.stack_size.unwrap());
         }
 
+        let _guard = handle.enter();
         let (store, _) = self.store.get_store()?;
-        let runtime = self
-            .wasi
-            .prepare_runtime(store.engine().clone(), &self.env, handle)?;
+        let runtime =
+            self.wasi
+                .prepare_runtime(store.engine().clone(), &self.env, handle.clone())?;
 
         // This is a slow operation, so let's temporarily wrap the runtime with
         // something that displays progress
-        let monitoring_runtime = MonitoringRuntime::new(runtime, pb.clone());
+        let monitoring_runtime = Arc::new(MonitoringRuntime::new(runtime, pb.clone()));
+        let runtime: Arc<dyn Runtime + Send + Sync> = monitoring_runtime.runtime.clone();
+        let monitoring_runtime: Arc<dyn Runtime + Send + Sync> = monitoring_runtime;
 
         let target = self.input.resolve_target(&monitoring_runtime, &pb)?;
 
         pb.finish_and_clear();
 
-        let runtime: Arc<dyn Runtime + Send + Sync> = Arc::new(monitoring_runtime.runtime);
-
         let result = {
             match target {
                 ExecutableTarget::WebAssembly { module, path } => {
@@ -166,7 +166,7 @@ impl Run {
             .get_command(id)
             .with_context(|| format!("Unable to get metadata for the \"{id}\" command"))?;
 
-        let uses = self.load_injected_packages(&*runtime)?;
+        let uses = self.load_injected_packages(&runtime)?;
 
         if WcgiRunner::can_run_command(cmd.metadata())? {
             self.run_wcgi(id, pkg, uses, runtime)
@@ -183,16 +183,25 @@ impl Run {
     }
 
     #[tracing::instrument(level = "debug", skip_all)]
-    fn load_injected_packages(&self, runtime: &dyn Runtime) -> Result<Vec<BinaryPackage>, Error> {
+    fn load_injected_packages(
+        &self,
+        runtime: &Arc<dyn Runtime + Send + Sync>,
+    ) -> Result<Vec<BinaryPackage>, Error> {
         let mut dependencies = Vec::new();
 
         for name in &self.wasi.uses {
             let specifier = PackageSpecifier::parse(name)
                 .with_context(|| format!("Unable to parse \"{name}\" as a package specifier"))?;
-            let pkg = runtime
-                .task_manager()
-                .block_on(BinaryPackage::from_registry(&specifier, runtime))
-                .with_context(|| format!("Unable to load \"{name}\""))?;
+            let pkg = {
+                let specifier = specifier.clone();
+                let inner_runtime = runtime.clone();
+                runtime
+                    .task_manager()
+                    .spawn_and_block_on(async move {
+                        BinaryPackage::from_registry(&specifier, inner_runtime.as_ref()).await
+                    })
+                    .with_context(|| format!("Unable to load \"{name}\""))?
+            };
             dependencies.push(pkg);
         }
 
@@ -461,7 +470,7 @@ impl PackageSource {
     /// internet.
     fn resolve_target(
         &self,
-        rt: &dyn Runtime,
+        rt: &Arc<dyn Runtime + Send + Sync>,
         pb: &ProgressBar,
     ) -> Result<ExecutableTarget, Error> {
         match self {
@@ -469,9 +478,11 @@ impl PackageSource {
             PackageSource::Dir(d) => ExecutableTarget::from_dir(d, rt, pb),
             PackageSource::Package(pkg) => {
                 pb.set_message("Loading from the registry");
-                let pkg = rt
-                    .task_manager()
-                    .block_on(BinaryPackage::from_registry(pkg, rt))?;
+                let inner_pck = pkg.clone();
+                let inner_rt = rt.clone();
+                let pkg = rt.task_manager().spawn_and_block_on(async move {
+                    BinaryPackage::from_registry(&inner_pck, inner_rt.as_ref()).await
+                })?;
                 Ok(ExecutableTarget::Package(pkg))
             }
         }
@@ -544,56 +555,44 @@ enum ExecutableTarget {
 impl ExecutableTarget {
     /// Try to load a Wasmer package from a directory containing a `wasmer.toml`
     /// file.
-    fn from_dir(dir: &Path, runtime: &dyn Runtime, pb: &ProgressBar) -> Result<Self, Error> {
+    #[tracing::instrument(level = "debug", skip_all)]
+    fn from_dir(
+        dir: &Path,
+        runtime: &Arc<dyn Runtime + Send + Sync>,
+        pb: &ProgressBar,
+    ) -> Result<Self, Error> {
         pb.set_message(format!("Loading \"{}\" into memory", dir.display()));
 
-        let webc = construct_webc_in_memory(dir)?;
-        let container = Container::from_bytes(webc)?;
+        let manifest_path = dir.join("wasmer.toml");
+        let webc = webc::wasmer_package::Package::from_manifest(manifest_path)?;
+        let container = Container::from(webc);
 
         pb.set_message("Resolving dependencies");
-        let pkg = runtime
-            .task_manager()
-            .block_on(BinaryPackage::from_webc(&container, runtime))?;
+        let inner_runtime = runtime.clone();
+        let pkg = runtime.task_manager().spawn_and_block_on(async move {
+            BinaryPackage::from_webc(&container, inner_runtime.as_ref()).await
+        })?;
 
         Ok(ExecutableTarget::Package(pkg))
     }
 
     /// Try to load a file into something that can be used to run it.
-    #[tracing::instrument(skip_all)]
-    fn from_file(path: &Path, runtime: &dyn Runtime, pb: &ProgressBar) -> Result<Self, Error> {
+    #[tracing::instrument(level = "debug", skip_all)]
+    fn from_file(
+        path: &Path,
+        runtime: &Arc<dyn Runtime + Send + Sync>,
+        pb: &ProgressBar,
+    ) -> Result<Self, Error> {
         pb.set_message(format!("Loading from \"{}\"", path.display()));
 
         match TargetOnDisk::from_file(path)? {
             TargetOnDisk::WebAssemblyBinary | TargetOnDisk::Wat => {
                 let wasm = std::fs::read(path)?;
-                let engine = runtime.engine().context("No engine available")?;
+
                 pb.set_message("Compiling to WebAssembly");
-
-                let tasks = runtime.task_manager();
-                let module_cache = runtime.module_cache();
-                let module_hash = ModuleHash::sha256(&wasm);
-
-                let module = match tasks.block_on(module_cache.load(module_hash, &engine)) {
-                    Ok(m) => m,
-                    Err(e) => {
-                        if !matches!(e, CacheError::NotFound) {
-                            tracing::warn!(
-                                module.path=%path.display(),
-                                module.hash=%module_hash,
-                                error=&e as &dyn std::error::Error,
-                                "Unable to deserialize the pre-compiled module from the module cache",
-                            );
-                        }
-
-                        let module = tracing::debug_span!("compiling_wasm")
-                            .in_scope(|| Module::new(&engine, &wasm))
-                            .with_context(|| format!("Unable to compile \"{}\"", path.display()))?;
-
-                        tasks.block_on(module_cache.save(module_hash, &engine, &module))?;
-
-                        module
-                    }
-                };
+                let module = runtime
+                    .load_module_sync(&wasm)
+                    .with_context(|| format!("Unable to compile \"{}\"", path.display()))?;
 
                 Ok(ExecutableTarget::WebAssembly {
                     module,
@@ -601,7 +600,7 @@ impl ExecutableTarget {
                 })
             }
             TargetOnDisk::Artifact => {
-                let engine = runtime.engine().context("No engine available")?;
+                let engine = runtime.engine();
                 pb.set_message("Deserializing pre-compiled WebAssembly module");
                 let module = unsafe { Module::deserialize_from_file(&engine, path)? };
 
@@ -613,58 +612,17 @@ impl ExecutableTarget {
             TargetOnDisk::LocalWebc => {
                 let container = Container::from_disk(path)?;
                 pb.set_message("Resolving dependencies");
-                let pkg = runtime
-                    .task_manager()
-                    .block_on(BinaryPackage::from_webc(&container, runtime))?;
+
+                let inner_runtime = runtime.clone();
+                let pkg = runtime.task_manager().spawn_and_block_on(async move {
+                    BinaryPackage::from_webc(&container, inner_runtime.as_ref()).await
+                })?;
                 Ok(ExecutableTarget::Package(pkg))
             }
         }
     }
 }
 
-#[tracing::instrument(level = "debug", skip_all)]
-fn construct_webc_in_memory(dir: &Path) -> Result<Vec<u8>, Error> {
-    let mut files = BTreeMap::new();
-    load_files_from_disk(&mut files, dir, dir)?;
-
-    let wasmer_toml = DirOrFile::File("wasmer.toml".into());
-    if let Some(toml_data) = files.remove(&wasmer_toml) {
-        // HACK(Michael-F-Bryan): The version of wapm-targz-to-pirita we are
-        // using doesn't know we renamed "wapm.toml" to "wasmer.toml", so we
-        // manually patch things up if people have already migrated their
-        // projects.
-        files
-            .entry(DirOrFile::File("wapm.toml".into()))
-            .or_insert(toml_data);
-    }
-
-    let functions = TransformManifestFunctions::default();
-    let webc = wapm_targz_to_pirita::generate_webc_file(files, dir, &functions)?;
-
-    Ok(webc)
-}
-
-fn load_files_from_disk(files: &mut FileMap, dir: &Path, base: &Path) -> Result<(), Error> {
-    let entries = dir
-        .read_dir()
-        .with_context(|| format!("Unable to read the contents of \"{}\"", dir.display()))?;
-
-    for entry in entries {
-        let path = entry?.path();
-        let relative_path = path.strip_prefix(base)?.to_path_buf();
-
-        if path.is_dir() {
-            load_files_from_disk(files, &path, base)?;
-            files.insert(DirOrFile::Dir(relative_path), Vec::new());
-        } else if path.is_file() {
-            let data = std::fs::read(&path)
-                .with_context(|| format!("Unable to read \"{}\"", path.display()))?;
-            files.insert(DirOrFile::File(relative_path), data);
-        }
-    }
-    Ok(())
-}
-
 #[cfg(feature = "coredump")]
 fn generate_coredump(err: &Error, source_name: String, coredump_path: &Path) -> Result<(), Error> {
     let err: &wasmer::RuntimeError = match err.downcast_ref() {
@@ -789,13 +747,16 @@ fn get_exit_code(
 
 #[derive(Debug)]
 struct MonitoringRuntime<R> {
-    runtime: R,
+    runtime: Arc<R>,
     progress: ProgressBar,
 }
 
 impl<R> MonitoringRuntime<R> {
     fn new(runtime: R, progress: ProgressBar) -> Self {
-        MonitoringRuntime { runtime, progress }
+        MonitoringRuntime {
+            runtime: Arc::new(runtime),
+            progress,
+        }
     }
 }
 
@@ -832,7 +793,7 @@ impl<R: wasmer_wasix::Runtime + Send + Sync> wasmer_wasix::Runtime for Monitorin
         })
     }
 
-    fn engine(&self) -> Option<wasmer::Engine> {
+    fn engine(&self) -> wasmer::Engine {
         self.runtime.engine()
     }
 
