@@ -1,3 +1,5 @@
+#![cfg_attr(docsrs, feature(doc_cfg, doc_auto_cfg))]
+
 use std::fmt;
 use std::mem::MaybeUninit;
 use std::net::IpAddr;
@@ -6,13 +8,14 @@ use std::net::Ipv6Addr;
 use std::net::Shutdown;
 use std::net::SocketAddr;
 use std::sync::Arc;
-use std::task::Context;
-use std::task::Poll;
 use std::time::Duration;
 use thiserror::Error;
 
 pub use bytes::Bytes;
 pub use bytes::BytesMut;
+pub use virtual_io::{handler_into_waker, InterestHandler};
+#[cfg(feature = "host-net")]
+pub use virtual_io::{InterestGuard, InterestHandlerWaker, InterestType};
 
 pub type Result<T> = std::result::Result<T, NetworkError>;
 
@@ -32,6 +35,12 @@ pub struct IpRoute {
     pub expires_at: Option<Duration>,
 }
 
+/// Represents an IO source
+pub trait VirtualIoSource: fmt::Debug + Send + Sync + 'static {
+    /// Removes a previously registered waker using a token
+    fn remove_handler(&mut self);
+}
+
 /// An implementation of virtual networking
 #[async_trait::async_trait]
 #[allow(unused_variables)]
@@ -172,21 +181,13 @@ pub trait VirtualNetworking: fmt::Debug + Send + Sync + 'static {
 
 pub type DynVirtualNetworking = Arc<dyn VirtualNetworking>;
 
-pub trait VirtualTcpListener: fmt::Debug + Send + Sync + 'static {
+pub trait VirtualTcpListener: VirtualIoSource + fmt::Debug + Send + Sync + 'static {
     /// Tries to accept a new connection
-    fn try_accept(&mut self) -> Option<Result<(Box<dyn VirtualTcpSocket + Sync>, SocketAddr)>>;
+    fn try_accept(&mut self) -> Result<(Box<dyn VirtualTcpSocket + Sync>, SocketAddr)>;
 
-    /// Polls the socket for new connections
-    fn poll_accept(
-        &mut self,
-        cx: &mut std::task::Context<'_>,
-    ) -> std::task::Poll<Result<(Box<dyn VirtualTcpSocket + Sync>, SocketAddr)>>;
-
-    /// Polls the socket for when there is data to be received
-    fn poll_accept_ready(
-        &mut self,
-        cx: &mut std::task::Context<'_>,
-    ) -> std::task::Poll<Result<usize>>;
+    /// Registers a waker for when a new connection has arrived. This uses
+    /// a stack machine which means more than one waker can be registered
+    fn set_handler(&mut self, handler: Box<dyn InterestHandler + Send + Sync>) -> Result<()>;
 
     /// Returns the local address of this TCP listener
     fn addr_local(&self) -> Result<SocketAddr>;
@@ -198,7 +199,7 @@ pub trait VirtualTcpListener: fmt::Debug + Send + Sync + 'static {
     fn ttl(&self) -> Result<u8>;
 }
 
-pub trait VirtualSocket: fmt::Debug + Send + Sync + 'static {
+pub trait VirtualSocket: VirtualIoSource + fmt::Debug + Send + Sync + 'static {
     /// Sets how many network hops the packets are permitted for new connections
     fn set_ttl(&mut self, ttl: u32) -> Result<()>;
 
@@ -211,17 +212,10 @@ pub trait VirtualSocket: fmt::Debug + Send + Sync + 'static {
     /// Returns the status/state of the socket
     fn status(&self) -> Result<SocketStatus>;
 
-    /// Polls the socket for when there is data to be received
-    fn poll_read_ready(
-        &mut self,
-        cx: &mut std::task::Context<'_>,
-    ) -> std::task::Poll<Result<usize>>;
-
-    /// Polls the socket for when the backpressure allows for writing to the socket
-    fn poll_write_ready(
-        &mut self,
-        cx: &mut std::task::Context<'_>,
-    ) -> std::task::Poll<Result<usize>>;
+    /// Registers a waker for when this connection is ready to receive
+    /// more data. Uses a stack machine which means more than one waker
+    /// can be registered
+    fn set_handler(&mut self, handler: Box<dyn InterestHandler + Send + Sync>) -> Result<()>;
 }
 
 #[derive(Debug, Copy, Clone, PartialEq, Eq)]
@@ -255,50 +249,23 @@ pub trait VirtualConnectedSocket: VirtualSocket + fmt::Debug + Send + Sync + 'st
     /// Tries to send out a datagram or stream of bytes on this socket
     fn try_send(&mut self, data: &[u8]) -> Result<usize>;
 
-    /// Sends out a datagram or stream of bytes on this socket
-    fn poll_send(&mut self, cx: &mut Context<'_>, data: &[u8]) -> Poll<Result<usize>>;
-
-    /// Attempts to flush the object, ensuring that any buffered data reach
-    /// their destination.
-    fn poll_flush(&mut self, cx: &mut Context<'_>) -> Poll<Result<()>>;
+    // Tries to flush any data in the local buffers
+    fn try_flush(&mut self) -> Result<()>;
 
     /// Closes the socket
     fn close(&mut self) -> Result<()>;
 
-    /// Recv a packet from the socket
-    fn poll_recv(
-        &mut self,
-        cx: &mut Context<'_>,
-        buf: &mut [MaybeUninit<u8>],
-    ) -> Poll<Result<usize>>;
-
-    /// Recv a packet from the socket
+    /// Tries to read a packet from the socket
     fn try_recv(&mut self, buf: &mut [MaybeUninit<u8>]) -> Result<usize>;
 }
 
 /// Connectionless sockets are able to send and receive datagrams and stream
 /// bytes to multiple addresses at the same time (peer-to-peer)
 pub trait VirtualConnectionlessSocket: VirtualSocket + fmt::Debug + Send + Sync + 'static {
-    /// Sends out a datagram or stream of bytes on this socket
-    /// to a specific address
-    fn poll_send_to(
-        &mut self,
-        cx: &mut Context<'_>,
-        data: &[u8],
-        addr: SocketAddr,
-    ) -> Poll<Result<usize>>;
-
     /// Sends out a datagram or stream of bytes on this socket
     /// to a specific address
     fn try_send_to(&mut self, data: &[u8], addr: SocketAddr) -> Result<usize>;
 
-    /// Recv a packet from the socket
-    fn poll_recv_from(
-        &mut self,
-        cx: &mut Context<'_>,
-        buf: &mut [MaybeUninit<u8>],
-    ) -> Poll<Result<(usize, SocketAddr)>>;
-
     /// Recv a packet from the socket
     fn try_recv_from(&mut self, buf: &mut [MaybeUninit<u8>]) -> Result<(usize, SocketAddr)>;
 }
@@ -311,27 +278,13 @@ pub trait VirtualIcmpSocket:
 }
 
 pub trait VirtualRawSocket: VirtualSocket + fmt::Debug + Send + Sync + 'static {
-    /// Sends out a datagram or stream of bytes on this socket
-    fn poll_send(&mut self, cx: &mut Context<'_>, data: &[u8]) -> Poll<Result<usize>>;
-
     /// Sends out a datagram or stream of bytes on this socket
     fn try_send(&mut self, data: &[u8]) -> Result<usize>;
 
-    /// Attempts to flush the object, ensuring that any buffered data reach
-    /// their destination.
-    fn poll_flush(&mut self, cx: &mut Context<'_>) -> Poll<Result<()>>;
-
     /// Attempts to flush the object, ensuring that any buffered data reach
     /// their destination.
     fn try_flush(&mut self) -> Result<()>;
 
-    /// Recv a packet from the socket
-    fn poll_recv(
-        &mut self,
-        cx: &mut Context<'_>,
-        buf: &mut [MaybeUninit<u8>],
-    ) -> Poll<Result<usize>>;
-
     /// Recv a packet from the socket
     fn try_recv(&mut self, buf: &mut [MaybeUninit<u8>]) -> Result<usize>;
 
