@@ -38,8 +38,7 @@ pub fn main() {
 }
 
 pub const DEFAULT_BOOT_WEBC: &str = "sharrattj/bash";
-//pub const DEFAULT_BOOT_WEBC: &str = "sharrattj/dash";
-pub const DEFAULT_BOOT_USES: [&str; 2] = ["sharrattj/coreutils", "sharrattj/catsay"];
+pub const DEFAULT_BOOT_USES: [&str; 1] = ["sharrattj/coreutils"];
 
 #[wasm_bindgen]
 pub fn start() -> Result<(), JsValue> {
@@ -72,6 +71,11 @@ pub fn start() -> Result<(), JsValue> {
     );
 
     let window = web_sys::window().unwrap();
+
+    // HACK: Make the xterm terminal publicly accessible so integration tests
+    // can tap into it
+    js_sys::Reflect::set(&window, &JsValue::from_str("xterm"), &terminal)?;
+
     let location = window.location().href().unwrap();
 
     let user_agent = USER_AGENT.clone();
