@@ -10,7 +10,7 @@ use http::{HeaderMap, StatusCode};
 use js_sys::Promise;
 use tokio::{
     io::{AsyncRead, AsyncSeek, AsyncWrite},
-    runtime::{Builder, Handle, Runtime},
+    runtime::{Builder, Runtime},
     sync::mpsc,
 };
 #[allow(unused_imports, dead_code)]
@@ -21,7 +21,7 @@ use wasmer_wasix::{
     http::{DynHttpClient, HttpRequest, HttpResponse},
     os::{TtyBridge, TtyOptions},
     runtime::{
-        module_cache::{FileSystemCache, ModuleCache, ThreadLocalCache},
+        module_cache::ModuleCache,
         package_loader::{BuiltinPackageLoader, PackageLoader},
         resolver::{Source, WapmSource},
         task_manager::TaskWasm,
@@ -37,6 +37,7 @@ use super::webgl::WebGl;
 #[cfg(feature = "webgl")]
 use super::webgl::WebGlCommand;
 use super::{common::*, pool::WebThreadPool};
+use crate::module_cache::WebWorkerModuleCache;
 
 #[derive(Debug)]
 pub(crate) enum TerminalCommandRx {
@@ -71,10 +72,7 @@ impl WebRuntime {
         let webgl_tx = GlContext::init(webgl2);
 
         let runtime: Arc<Runtime> = Arc::new(Builder::new_current_thread().build().unwrap());
-        let runtime = Arc::new(WebTaskManager {
-            pool: pool.clone(),
-            runtime,
-        });
+        let runtime = Arc::new(WebTaskManager { pool: pool.clone() });
 
         let http_client = Arc::new(WebHttpClient { pool: pool.clone() });
         let source = WapmSource::new(
@@ -88,13 +86,8 @@ impl WebRuntime {
         // even if the filesystem cache fails (i.e. because we're compiled to
         // wasm32-unknown-unknown and running in a browser), the in-memory layer
         // should still work.
-        let package_loader = BuiltinPackageLoader::new_with_client(
-            BuiltinPackageLoader::default_cache_dir(&wasmer_dir),
-            http_client.clone(),
-        );
-        let module_cache = ThreadLocalCache::default().with_fallback(FileSystemCache::new(
-            FileSystemCache::default_cache_dir(&wasmer_dir),
-        ));
+        let package_loader = BuiltinPackageLoader::new_only_client(http_client.clone());
+        let module_cache = WebWorkerModuleCache::default();
         WebRuntime {
             pool,
             tasks: runtime,
@@ -118,7 +111,6 @@ impl VirtualNetworking for WebVirtualNetworking {}
 #[derive(Debug, Clone)]
 pub(crate) struct WebTaskManager {
     pool: WebThreadPool,
-    runtime: Arc<Runtime>,
 }
 
 struct WebRuntimeGuard<'g> {
@@ -179,19 +171,6 @@ impl VirtualTaskManager for WebTaskManager {
         Ok(())
     }
 
-    /// Returns a runtime that can be used for asynchronous tasks
-    fn runtime(&self) -> &Handle {
-        self.runtime.handle()
-    }
-
-    /// Enters a runtime context
-    #[allow(dyn_drop)]
-    fn runtime_enter<'g>(&'g self) -> Box<dyn std::ops::Drop + 'g> {
-        Box::new(WebRuntimeGuard {
-            inner: self.runtime.enter(),
-        })
-    }
-
     /// Starts an asynchronous task will will run on a dedicated thread
     /// pulled from the worker pool that has a stateful thread local variable
     /// It is ok for this task to block execution and any async futures within its scope
@@ -314,8 +293,8 @@ impl VirtualFile for TermStdout {
         Ok(())
     }
 
-    fn unlink(&mut self) -> wasmer_wasix::virtual_fs::Result<()> {
-        Ok(())
+    fn unlink(&mut self) -> BoxFuture<'static, wasmer_wasix::virtual_fs::Result<()>> {
+        Box::pin(async { Ok(()) })
     }
 
     fn poll_read_ready(
@@ -427,8 +406,8 @@ impl VirtualFile for TermLog {
         Ok(())
     }
 
-    fn unlink(&mut self) -> wasmer_wasix::virtual_fs::Result<()> {
-        Ok(())
+    fn unlink(&mut self) -> BoxFuture<'static, wasmer_wasix::virtual_fs::Result<()>> {
+        Box::pin(async { Ok(()) })
     }
 
     fn poll_read_ready(
@@ -569,7 +548,9 @@ impl wasmer_wasix::http::HttpClient for WebHttpClient {
         self.pool.spawn_shared(Box::new(move || {
             Box::pin(async move {
                 let res = Self::do_request(request).await;
-                let _ = tx.send(res);
+                if let Err(err) = tx.send(res) {
+                    tracing::error!("failed to reply http response to caller - {:?}", err);
+                }
             })
         }));
         Box::pin(async move { rx.await.unwrap() })
