@@ -11,10 +11,11 @@ use std::{
 use futures::future::BoxFuture;
 use tokio::io::{AsyncRead, AsyncSeek, AsyncWrite};
 use virtual_fs::{FsError, Pipe as VirtualPipe, VirtualFile};
+use virtual_io::{InterestType, StatefulHandler};
 use virtual_net::NetworkError;
 use wasmer_wasix_types::{
     types::Eventtype,
-    wasi,
+    wasi::{self, EpollType},
     wasi::{Errno, EventFdReadwrite, Eventrwflags, Subscription},
 };
 
@@ -26,6 +27,7 @@ use crate::{
     utils::{OwnedRwLockReadGuard, OwnedRwLockWriteGuard},
 };
 
+#[derive(Debug, Clone)]
 pub(crate) enum InodeValFilePollGuardMode {
     File(Arc<RwLock<Box<dyn VirtualFile + Send + Sync + 'static>>>),
     EventNotifications(Arc<NotificationInner>),
@@ -33,7 +35,7 @@ pub(crate) enum InodeValFilePollGuardMode {
     Pipe { pipe: Arc<RwLock<Box<VirtualPipe>>> },
 }
 
-pub(crate) struct InodeValFilePollGuard {
+pub struct InodeValFilePollGuard {
     pub(crate) fd: u32,
     pub(crate) peb: PollEventSet,
     pub(crate) subscription: Subscription,
@@ -48,17 +50,17 @@ impl InodeValFilePollGuard {
         guard: &Kind,
     ) -> Option<Self> {
         let mode = match guard.deref() {
-            Kind::EventNotifications(inner) => {
+            Kind::EventNotifications { inner, .. } => {
                 InodeValFilePollGuardMode::EventNotifications(inner.clone())
             }
-            Kind::Socket { socket } => InodeValFilePollGuardMode::Socket {
+            Kind::Socket { socket, .. } => InodeValFilePollGuardMode::Socket {
                 inner: socket.inner.clone(),
             },
             Kind::File {
                 handle: Some(handle),
                 ..
             } => InodeValFilePollGuardMode::File(handle.clone()),
-            Kind::Pipe { pipe } => InodeValFilePollGuardMode::Pipe {
+            Kind::Pipe { pipe, .. } => InodeValFilePollGuardMode::Pipe {
                 pipe: Arc::new(RwLock::new(Box::new(pipe.clone()))),
             },
             _ => {
@@ -116,11 +118,13 @@ impl std::fmt::Debug for InodeValFilePollGuard {
     }
 }
 
-pub(crate) struct InodeValFilePollGuardJoin {
+#[derive(Debug)]
+pub struct InodeValFilePollGuardJoin {
     mode: InodeValFilePollGuardMode,
     fd: u32,
     peb: PollEventSet,
     subscription: Subscription,
+    spent: bool,
 }
 
 impl InodeValFilePollGuardJoin {
@@ -130,6 +134,7 @@ impl InodeValFilePollGuardJoin {
             fd: guard.fd,
             peb: guard.peb,
             subscription: guard.subscription,
+            spent: false,
         }
     }
     pub(crate) fn fd(&self) -> u32 {
@@ -138,12 +143,29 @@ impl InodeValFilePollGuardJoin {
     pub(crate) fn peb(&self) -> PollEventSet {
         self.peb
     }
+    pub fn is_spent(&self) -> bool {
+        self.spent
+    }
+    pub fn reset(&mut self) {
+        match &self.mode {
+            InodeValFilePollGuardMode::File(_) => {}
+            InodeValFilePollGuardMode::EventNotifications(inner) => {
+                inner.reset();
+            }
+            InodeValFilePollGuardMode::Socket { .. } => {}
+            InodeValFilePollGuardMode::Pipe { .. } => {}
+        }
+        self.spent = false;
+    }
 }
 
+pub const POLL_GUARD_MAX_RET: usize = 4;
+
 impl Future for InodeValFilePollGuardJoin {
-    type Output = heapless::Vec<EventResult, 4>;
+    type Output = heapless::Vec<(EventResult, EpollType), POLL_GUARD_MAX_RET>;
 
     fn poll(mut self: Pin<&mut Self>, cx: &mut std::task::Context<'_>) -> Poll<Self::Output> {
+        // Otherwise we need to register for the event
         let fd = self.fd();
         let waker = cx.waker();
         let mut has_read = false;
@@ -171,67 +193,6 @@ impl Future for InodeValFilePollGuardJoin {
                 }
             }
         }
-        if has_close {
-            let is_closed = match &mut self.mode {
-                InodeValFilePollGuardMode::File(file) => {
-                    let mut guard = file.write().unwrap();
-                    let file = Pin::new(guard.as_mut());
-                    file.poll_shutdown(cx).is_ready()
-                }
-                InodeValFilePollGuardMode::EventNotifications { .. } => false,
-                InodeValFilePollGuardMode::Socket { ref inner } => {
-                    let mut guard = inner.protected.write().unwrap();
-                    let is_closed = if has_read || has_write {
-                        // this will be handled in the read/write poll instead
-                        false
-                    } else {
-                        // we do a read poll which will error out if its closed
-                        #[allow(clippy::match_like_matches_macro)]
-                        match guard.poll_read_ready(cx) {
-                            Poll::Ready(Ok(0)) => true,
-                            Poll::Ready(Err(NetworkError::ConnectionAborted))
-                            | Poll::Ready(Err(NetworkError::ConnectionRefused))
-                            | Poll::Ready(Err(NetworkError::ConnectionReset))
-                            | Poll::Ready(Err(NetworkError::BrokenPipe))
-                            | Poll::Ready(Err(NetworkError::NotConnected))
-                            | Poll::Ready(Err(NetworkError::UnexpectedEof)) => true,
-                            _ => false,
-                        }
-                    };
-                    if is_closed {
-                        !replace(&mut guard.notifications.closed, true)
-                    } else {
-                        false
-                    }
-                }
-                InodeValFilePollGuardMode::Pipe { pipe } => {
-                    let mut guard = pipe.write().unwrap();
-                    let pipe = Pin::new(guard.as_mut());
-                    pipe.poll_shutdown(cx).is_ready()
-                }
-            };
-            if is_closed {
-                ret.push(EventResult {
-                    userdata: self.subscription.userdata,
-                    error: Errno::Success,
-                    type_: self.subscription.type_,
-                    inner: match self.subscription.type_ {
-                        Eventtype::FdRead | Eventtype::FdWrite => {
-                            EventResultType::Fd(EventFdReadwrite {
-                                nbytes: 0,
-                                flags: if has_hangup {
-                                    Eventrwflags::FD_READWRITE_HANGUP
-                                } else {
-                                    Eventrwflags::empty()
-                                },
-                            })
-                        }
-                        Eventtype::Clock => EventResultType::Clock(0),
-                    },
-                })
-                .ok();
-            }
-        }
         if has_read {
             let poll_result = match &mut self.mode {
                 InodeValFilePollGuardMode::File(file) => {
@@ -242,32 +203,37 @@ impl Future for InodeValFilePollGuardJoin {
                 InodeValFilePollGuardMode::EventNotifications(inner) => inner.poll(waker).map(Ok),
                 InodeValFilePollGuardMode::Socket { ref inner } => {
                     let mut guard = inner.protected.write().unwrap();
-                    let res = guard.poll_read_ready(cx).map_err(net_error_into_io_err);
-                    match res {
-                        Poll::Ready(Err(err)) if is_err_closed(&err) => {
-                            tracing::trace!("socket read ready error (fd={}) - {}", fd, err);
-                            if !replace(&mut guard.notifications.closed, true) {
-                                Poll::Ready(Ok(0))
-                            } else {
+                    if guard.handler_state.take(InterestType::Readable) {
+                        Poll::Ready(Ok(8192))
+                    } else {
+                        let handler =
+                            StatefulHandler::new(cx.waker().into(), guard.handler_state.clone());
+
+                        let res = guard
+                            .add_handler(handler, InterestType::Readable)
+                            .map_err(net_error_into_io_err);
+                        match res {
+                            Err(err) if is_err_closed(&err) => {
+                                tracing::trace!("socket read ready error (fd={}) - {}", fd, err);
+                                if !replace(&mut guard.notifications.closed, true) {
+                                    Poll::Ready(Ok(0))
+                                } else {
+                                    Poll::Pending
+                                }
+                            }
+                            Err(err) => {
+                                tracing::debug!("poll socket error - {}", err);
+                                if !replace(&mut guard.notifications.failed, true) {
+                                    Poll::Ready(Ok(0))
+                                } else {
+                                    Poll::Pending
+                                }
+                            }
+                            Ok(()) => {
+                                drop(guard);
                                 Poll::Pending
                             }
                         }
-                        Poll::Ready(Err(err)) => {
-                            tracing::debug!("poll socket error - {}", err);
-                            if !replace(&mut guard.notifications.failed, true) {
-                                Poll::Ready(Ok(0))
-                            } else {
-                                Poll::Pending
-                            }
-                        }
-                        Poll::Ready(Ok(amt)) => {
-                            if guard.notifications.closed {
-                                Poll::Pending
-                            } else {
-                                Poll::Ready(Ok(amt))
-                            }
-                        }
-                        Poll::Pending => Poll::Pending,
                     }
                 }
                 InodeValFilePollGuardMode::Pipe { pipe } => {
@@ -278,25 +244,32 @@ impl Future for InodeValFilePollGuardJoin {
             };
             match poll_result {
                 Poll::Ready(Err(err)) if has_close && is_err_closed(&err) => {
-                    ret.push(EventResult {
-                        userdata: self.subscription.userdata,
-                        error: Errno::Success,
-                        type_: self.subscription.type_,
-                        inner: match self.subscription.type_ {
-                            Eventtype::FdRead | Eventtype::FdWrite => {
-                                EventResultType::Fd(EventFdReadwrite {
-                                    nbytes: 0,
-                                    flags: if has_hangup {
-                                        Eventrwflags::FD_READWRITE_HANGUP
-                                    } else {
-                                        Eventrwflags::empty()
-                                    },
-                                })
-                            }
-                            Eventtype::Clock => EventResultType::Clock(0),
-                        },
-                    })
-                    .ok();
+                    let inner = match self.subscription.type_ {
+                        Eventtype::FdRead | Eventtype::FdWrite => {
+                            Some(EventResultType::Fd(EventFdReadwrite {
+                                nbytes: 0,
+                                flags: if has_hangup {
+                                    Eventrwflags::FD_READWRITE_HANGUP
+                                } else {
+                                    Eventrwflags::empty()
+                                },
+                            }))
+                        }
+                        Eventtype::Clock => Some(EventResultType::Clock(0)),
+                        Eventtype::Unknown => None,
+                    };
+                    if let Some(inner) = inner {
+                        ret.push((
+                            EventResult {
+                                userdata: self.subscription.userdata,
+                                error: Errno::Success,
+                                type_: self.subscription.type_,
+                                inner,
+                            },
+                            EpollType::EPOLLHUP,
+                        ))
+                        .ok();
+                    }
                 }
                 Poll::Ready(bytes_available) => {
                     let mut error = Errno::Success;
@@ -307,25 +280,36 @@ impl Future for InodeValFilePollGuardJoin {
                             0
                         }
                     };
-                    ret.push(EventResult {
-                        userdata: self.subscription.userdata,
-                        error,
-                        type_: self.subscription.type_,
-                        inner: match self.subscription.type_ {
-                            Eventtype::FdRead | Eventtype::FdWrite => {
-                                EventResultType::Fd(EventFdReadwrite {
-                                    nbytes: bytes_available as u64,
-                                    flags: if bytes_available == 0 {
-                                        Eventrwflags::FD_READWRITE_HANGUP
-                                    } else {
-                                        Eventrwflags::empty()
-                                    },
-                                })
-                            }
-                            Eventtype::Clock => EventResultType::Clock(0),
-                        },
-                    })
-                    .ok();
+                    let inner = match self.subscription.type_ {
+                        Eventtype::FdRead | Eventtype::FdWrite => {
+                            Some(EventResultType::Fd(EventFdReadwrite {
+                                nbytes: bytes_available as u64,
+                                flags: if bytes_available == 0 {
+                                    Eventrwflags::FD_READWRITE_HANGUP
+                                } else {
+                                    Eventrwflags::empty()
+                                },
+                            }))
+                        }
+                        Eventtype::Clock => Some(EventResultType::Clock(0)),
+                        Eventtype::Unknown => None,
+                    };
+                    if let Some(inner) = inner {
+                        ret.push((
+                            EventResult {
+                                userdata: self.subscription.userdata,
+                                error,
+                                type_: self.subscription.type_,
+                                inner,
+                            },
+                            if error == Errno::Success {
+                                EpollType::EPOLLIN
+                            } else {
+                                EpollType::EPOLLERR
+                            },
+                        ))
+                        .ok();
+                    }
                 }
                 Poll::Pending => {}
             };
@@ -340,32 +324,41 @@ impl Future for InodeValFilePollGuardJoin {
                 InodeValFilePollGuardMode::EventNotifications(inner) => inner.poll(waker).map(Ok),
                 InodeValFilePollGuardMode::Socket { ref inner } => {
                     let mut guard = inner.protected.write().unwrap();
-                    let res = guard.poll_write_ready(cx).map_err(net_error_into_io_err);
-                    match res {
-                        Poll::Ready(Err(err)) if is_err_closed(&err) => {
-                            tracing::trace!("socket write ready error (fd={}) - err={}", fd, err);
-                            if !replace(&mut guard.notifications.closed, true) {
-                                Poll::Ready(Ok(0))
-                            } else {
+                    if guard.handler_state.take(InterestType::Writable) {
+                        Poll::Ready(Ok(8192))
+                    } else {
+                        let handler =
+                            StatefulHandler::new(cx.waker().into(), guard.handler_state.clone());
+
+                        let res = guard
+                            .add_handler(handler, InterestType::Writable)
+                            .map_err(net_error_into_io_err);
+                        match res {
+                            Err(err) if is_err_closed(&err) => {
+                                tracing::trace!(
+                                    "socket write ready error (fd={}) - err={}",
+                                    fd,
+                                    err
+                                );
+                                if !replace(&mut guard.notifications.closed, true) {
+                                    Poll::Ready(Ok(0))
+                                } else {
+                                    Poll::Pending
+                                }
+                            }
+                            Err(err) => {
+                                tracing::debug!("poll socket error - {}", err);
+                                if !replace(&mut guard.notifications.failed, true) {
+                                    Poll::Ready(Ok(0))
+                                } else {
+                                    Poll::Pending
+                                }
+                            }
+                            Ok(()) => {
+                                drop(guard);
                                 Poll::Pending
                             }
                         }
-                        Poll::Ready(Err(err)) => {
-                            tracing::debug!("poll socket error - {}", err);
-                            if !replace(&mut guard.notifications.failed, true) {
-                                Poll::Ready(Ok(0))
-                            } else {
-                                Poll::Pending
-                            }
-                        }
-                        Poll::Ready(Ok(amt)) => {
-                            if guard.notifications.closed {
-                                Poll::Pending
-                            } else {
-                                Poll::Ready(Ok(amt))
-                            }
-                        }
-                        Poll::Pending => Poll::Pending,
                     }
                 }
                 InodeValFilePollGuardMode::Pipe { pipe } => {
@@ -376,25 +369,32 @@ impl Future for InodeValFilePollGuardJoin {
             };
             match poll_result {
                 Poll::Ready(Err(err)) if has_close && is_err_closed(&err) => {
-                    ret.push(EventResult {
-                        userdata: self.subscription.userdata,
-                        error: Errno::Success,
-                        type_: self.subscription.type_,
-                        inner: match self.subscription.type_ {
-                            Eventtype::FdRead | Eventtype::FdWrite => {
-                                EventResultType::Fd(EventFdReadwrite {
-                                    nbytes: 0,
-                                    flags: if has_hangup {
-                                        Eventrwflags::FD_READWRITE_HANGUP
-                                    } else {
-                                        Eventrwflags::empty()
-                                    },
-                                })
-                            }
-                            Eventtype::Clock => EventResultType::Clock(0),
-                        },
-                    })
-                    .ok();
+                    let inner = match self.subscription.type_ {
+                        Eventtype::FdRead | Eventtype::FdWrite => {
+                            Some(EventResultType::Fd(EventFdReadwrite {
+                                nbytes: 0,
+                                flags: if has_hangup {
+                                    Eventrwflags::FD_READWRITE_HANGUP
+                                } else {
+                                    Eventrwflags::empty()
+                                },
+                            }))
+                        }
+                        Eventtype::Clock => Some(EventResultType::Clock(0)),
+                        Eventtype::Unknown => None,
+                    };
+                    if let Some(inner) = inner {
+                        ret.push((
+                            EventResult {
+                                userdata: self.subscription.userdata,
+                                error: Errno::Success,
+                                type_: self.subscription.type_,
+                                inner,
+                            },
+                            EpollType::EPOLLHUP,
+                        ))
+                        .ok();
+                    }
                 }
                 Poll::Ready(bytes_available) => {
                     let mut error = Errno::Success;
@@ -405,31 +405,42 @@ impl Future for InodeValFilePollGuardJoin {
                             0
                         }
                     };
-                    ret.push(EventResult {
-                        userdata: self.subscription.userdata,
-                        error,
-                        type_: self.subscription.type_,
-                        inner: match self.subscription.type_ {
-                            Eventtype::FdRead | Eventtype::FdWrite => {
-                                EventResultType::Fd(EventFdReadwrite {
-                                    nbytes: bytes_available as u64,
-                                    flags: if bytes_available == 0 {
-                                        Eventrwflags::FD_READWRITE_HANGUP
-                                    } else {
-                                        Eventrwflags::empty()
-                                    },
-                                })
-                            }
-                            Eventtype::Clock => EventResultType::Clock(0),
-                        },
-                    })
-                    .ok();
+                    let inner = match self.subscription.type_ {
+                        Eventtype::FdRead | Eventtype::FdWrite => {
+                            Some(EventResultType::Fd(EventFdReadwrite {
+                                nbytes: bytes_available as u64,
+                                flags: if bytes_available == 0 {
+                                    Eventrwflags::FD_READWRITE_HANGUP
+                                } else {
+                                    Eventrwflags::empty()
+                                },
+                            }))
+                        }
+                        Eventtype::Clock => Some(EventResultType::Clock(0)),
+                        Eventtype::Unknown => None,
+                    };
+                    if let Some(inner) = inner {
+                        ret.push((
+                            EventResult {
+                                userdata: self.subscription.userdata,
+                                error,
+                                type_: self.subscription.type_,
+                                inner,
+                            },
+                            if error == Errno::Success {
+                                EpollType::EPOLLOUT
+                            } else {
+                                EpollType::EPOLLERR
+                            },
+                        ))
+                        .ok();
+                    }
                 }
                 Poll::Pending => {}
             };
         }
-
         if !ret.is_empty() {
+            self.spent = true;
             return Poll::Ready(ret);
         }
         Poll::Pending
@@ -738,7 +749,7 @@ fn is_err_closed(err: &std::io::Error) -> bool {
         || err.kind() == std::io::ErrorKind::UnexpectedEof
 }
 
-fn net_error_into_io_err(net_error: NetworkError) -> std::io::Error {
+pub fn net_error_into_io_err(net_error: NetworkError) -> std::io::Error {
     use std::io::ErrorKind;
     match net_error {
         NetworkError::InvalidFd => ErrorKind::BrokenPipe.into(),
