@@ -7,14 +7,16 @@ use std::{
     collections::{HashMap, HashSet, VecDeque},
     ops::{Deref, DerefMut},
     path::{Component, Path, PathBuf},
+    pin::Pin,
     sync::{
         atomic::{AtomicBool, AtomicU32, AtomicU64, Ordering},
         Arc, Mutex, RwLock, Weak,
     },
+    task::{Context, Poll},
 };
 
 use crate::state::{Stderr, Stdin, Stdout};
-use futures::{future::BoxFuture, TryStreamExt};
+use futures::{future::BoxFuture, Future, TryStreamExt};
 #[cfg(feature = "enable-serde")]
 use serde_derive::{Deserialize, Serialize};
 use tokio::io::AsyncWriteExt;
@@ -28,10 +30,11 @@ use wasmer_wasix_types::{
     },
 };
 
-pub use self::fd::{Fd, InodeVal, Kind};
+pub use self::fd::{EpollFd, EpollInterest, EpollJoinGuard, Fd, InodeVal, Kind};
 pub(crate) use self::inode_guard::{
-    InodeValFilePollGuard, InodeValFilePollGuardJoin, InodeValFileReadGuard,
-    InodeValFileWriteGuard, WasiStateFileGuard,
+    net_error_into_io_err, InodeValFilePollGuard, InodeValFilePollGuardJoin,
+    InodeValFilePollGuardMode, InodeValFileReadGuard, InodeValFileWriteGuard, WasiStateFileGuard,
+    POLL_GUARD_MAX_RET,
 };
 pub use self::notification::NotificationInner;
 use crate::syscalls::map_io_err;
@@ -405,7 +408,6 @@ fn create_dir_all(fs: &dyn FileSystem, path: &Path) -> Result<(), virtual_fs::Fs
 pub struct WasiFs {
     //pub repo: Repo,
     pub preopen_fds: RwLock<Vec<u32>>,
-    pub name_map: HashMap<String, Inode>,
     pub fd_map: Arc<RwLock<HashMap<WasiFd, Fd>>>,
     pub next_fd: AtomicU32,
     pub current_dir: Mutex<String>,
@@ -437,7 +439,6 @@ impl WasiFs {
         let fd_map = self.fd_map.read().unwrap().clone();
         Self {
             preopen_fds: RwLock::new(self.preopen_fds.read().unwrap().clone()),
-            name_map: self.name_map.clone(),
             fd_map: Arc::new(RwLock::new(fd_map)),
             next_fd: AtomicU32::new(self.next_fd.load(Ordering::SeqCst)),
             current_dir: Mutex::new(self.current_dir.lock().unwrap().clone()),
@@ -713,7 +714,6 @@ impl WasiFs {
 
         let wasi_fs = Self {
             preopen_fds: RwLock::new(vec![]),
-            name_map: HashMap::new(),
             fd_map: Arc::new(RwLock::new(HashMap::new())),
             next_fd: AtomicU32::new(3),
             current_dir: Mutex::new("/".to_string()),
@@ -1231,7 +1231,8 @@ impl WasiFs {
                     Kind::File { .. }
                     | Kind::Socket { .. }
                     | Kind::Pipe { .. }
-                    | Kind::EventNotifications { .. } => {
+                    | Kind::EventNotifications { .. }
+                    | Kind::Epoll { .. } => {
                         return Err(Errno::Notdir);
                     }
                     Kind::Symlink {
@@ -1526,35 +1527,52 @@ impl WasiFs {
     pub async fn flush(&self, fd: WasiFd) -> Result<(), Errno> {
         match fd {
             __WASI_STDIN_FILENO => (),
-            __WASI_STDOUT_FILENO => WasiInodes::stdout_mut(&self.fd_map)
-                .map_err(fs_error_into_wasi_err)?
-                .flush()
-                .await
-                .map_err(map_io_err)?,
-            __WASI_STDERR_FILENO => WasiInodes::stderr_mut(&self.fd_map)
-                .map_err(fs_error_into_wasi_err)?
-                .flush()
-                .await
-                .map_err(map_io_err)?,
+            __WASI_STDOUT_FILENO => {
+                let mut file =
+                    WasiInodes::stdout_mut(&self.fd_map).map_err(fs_error_into_wasi_err)?;
+                file.flush().await.map_err(map_io_err)?
+            }
+            __WASI_STDERR_FILENO => {
+                let mut file =
+                    WasiInodes::stderr_mut(&self.fd_map).map_err(fs_error_into_wasi_err)?;
+                file.flush().await.map_err(map_io_err)?
+            }
             _ => {
                 let fd = self.get_fd(fd)?;
                 if fd.rights.contains(Rights::FD_DATASYNC) {
                     return Err(Errno::Access);
                 }
 
-                let guard = fd.inode.read();
-                match guard.deref() {
-                    Kind::File {
-                        handle: Some(file), ..
-                    } => {
-                        let mut file = file.write().unwrap();
-                        file.flush().await.map_err(|_| Errno::Io)?
+                let work = {
+                    let guard = fd.inode.read();
+                    match guard.deref() {
+                        Kind::File {
+                            handle: Some(file), ..
+                        } => {
+                            struct FlushPoller {
+                                file: Arc<RwLock<Box<dyn VirtualFile + Send + Sync>>>,
+                            }
+                            impl Future for FlushPoller {
+                                type Output = Result<(), Errno>;
+                                fn poll(
+                                    self: Pin<&mut Self>,
+                                    cx: &mut Context<'_>,
+                                ) -> Poll<Self::Output> {
+                                    let mut file = self.file.write().unwrap();
+                                    Pin::new(file.as_mut())
+                                        .poll_flush(cx)
+                                        .map_err(|_| Errno::Io)
+                                }
+                            }
+                            FlushPoller { file: file.clone() }
+                        }
+                        // TODO: verify this behavior
+                        Kind::Dir { .. } => return Err(Errno::Isdir),
+                        Kind::Buffer { .. } => return Ok(()),
+                        _ => return Err(Errno::Io),
                     }
-                    // TODO: verify this behavior
-                    Kind::Dir { .. } => return Err(Errno::Isdir),
-                    Kind::Buffer { .. } => (),
-                    _ => return Err(Errno::Io),
-                }
+                };
+                work.await?
             }
         }
         Ok(())
